/* 
 * Copyright (c) 2020 Brian Hume.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * For a copy of the GNU General Public License, see <http://www.gnu.org/licenses/>.
 */
/*
 * The purpose of this program is 2 fold, 
 * 1. I wanted to try my hand a writing a small piece of a firewall that I could actually put to use
 * 2. I wanted to be able to block access to RDP on my home server for all IP addresses except
 *    those who I knew or who could send in an access key
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>

#include <pthread.h>
#include <signal.h>

#include <ifaddrs.h>
#include <time.h>
#include <sys/select.h>

#define PFNRM  "\x1B[0m"
#define PFRED  "\x1B[0;31m"
#define PFGRN  "\x1B[0;32m"
#define PFYEL  "\x1B[1;33m"
#define PFBLU  "\x1B[0;34m"
#define PFMAG  "\x1B[0;35m"
#define PFCYN  "\x1B[0;36m"
#define PFWHT  "\x1B[0;37m"
#define PFLIM  "\x1B[1;32m"


/*
The plan
Read config file for Port for forwarding, 
dest IP+port, Port of updating and IP list of valid soruces

If we recieve a message on the updating port and that message matches spec then we add the ip address to a temp list (1hr) and log it
If we recieve a message from a valid ip we log the first message and pass it on, we only log 1 message per hour
If we recieve a message from an invalid IP we log it
If we recieve a message we can not understand we log what we can
*/

//Temp IPV4 addresses
typedef struct
{
	int endOfLife;
	unsigned int ipAddress;
} timeoutAddress;

//Parameters for link thread
typedef struct
{
	int socket;
	socklen_t addrlen;
	struct sockaddr_in addr;
} linkInit;

//Config struct
typedef struct
{
	unsigned short portIn;
	unsigned short portOut;
	unsigned int ipAddressOut;
	unsigned short portUpdate;
	unsigned int ipAddressList[10];
	unsigned int ipAddressCount;
} configValues;


const char ACCESSKEY[] = "fakekey";//move access key to config
const char HTTP_AFFIRM[] = "HTTP/1.1 200 OK\r\n\r\nHello friend\r\n\r\n";
configValues config;
int LocalIP = -1;
int fatalError = 0;
int linkTimeout = 60;
//fixme add a mutex
timeoutAddress tempAddresses[10];
int tempIndex = 0;

int linkCount = 0;

int AddressIsGood(struct sockaddr_in addr, socklen_t addrLen)
{
	int ipAddr = addr.sin_addr.s_addr;
	for(int itr = 0; itr < config.ipAddressCount; itr++)
	{
		if (ipAddr == config.ipAddressList[itr])
			return 1;
		if (ipAddr == tempAddresses[itr].ipAddress && tempAddresses[itr].endOfLife > time(0))
			return 1;
	}
	return 0;
}

void ipv4ToString(int ip, char string[INET6_ADDRSTRLEN])
{
	if (ip == 0xffffffff)
		snprintf(string, INET6_ADDRSTRLEN, "BROADCAST");
	else if (ip & 0x000000ff >= 0x000000EB && ip & 0x000000ff < 0x000000F0)
		snprintf(string, INET6_ADDRSTRLEN, "MULTICAST");
	else if (ip & 0x000000ff == 0x00000000)
		snprintf(string, INET6_ADDRSTRLEN, "THIS HOST");
	else if (ip & 0x000000ff == 0x0000007f)
		snprintf(string, INET6_ADDRSTRLEN, "LOOP BACK");
	else
		inet_ntop(AF_INET, (void *)&ip, string, INET6_ADDRSTRLEN);
}

int ReadConfig()
{
	//variables
	int itr;
	char ipAddressList[10][INET6_ADDRSTRLEN];
	char outAddress[INET6_ADDRSTRLEN];
	memset(&config, 0, sizeof(config));
	memset(&tempAddresses, 0, sizeof(tempAddresses));

	//Open config file
	FILE *configFile = fopen("./config", "r");
	if (configFile == 0)
	{
		printf(PFRED "error reading config file\n" PFNRM);
		return -1;
	}

	//Read config file
	fscanf(configFile, "in-port %hu\n", &config.portIn);
	fscanf(configFile, "out-port %hu\n", &config.portOut);
	fscanf(configFile, "out-ip %s\n", &outAddress);
	fscanf(configFile, "update-port %hu\n", &config.portUpdate);
	for (itr = 0; itr < 10 && !feof(configFile) ; itr++)
		if (fscanf(configFile, "ip-address %s\n", ipAddressList[itr]) == 0)
			break;
	config.ipAddressCount = itr;

	inet_pton(AF_INET, outAddress, &config.ipAddressOut);
	for (itr = 0; itr < config.ipAddressCount; itr++)
		inet_pton(AF_INET, ipAddressList[itr], &config.ipAddressList[itr]);

	return 0;
}

void getMyIPAddress()
{
	struct ifaddrs *ifAddress, *tmp;
	if (getifaddrs(&ifAddress) != 0)
	{
		printf(PFRED "packet-helper Failed to get eth0 ip address %d\n" PFNRM, time(0));
		exit(-1);
	}
	tmp = ifAddress;
	while(tmp)
	{
		if (strcmp(tmp->ifa_name, "eth0") == 0 && tmp->ifa_addr->sa_family == AF_INET)
		{
			LocalIP = ((struct sockaddr_in*)tmp->ifa_addr)->sin_addr.s_addr;
			char ipString[INET6_ADDRSTRLEN];
			ipv4ToString(LocalIP, ipString);
			printf(PFGRN "packet-helper Got my eth0 ip address " PFLIM " %s " PFGRN " %d\n" PFNRM, ipString, time(0));
			return;
		}
		tmp = tmp->ifa_next;
	}

	printf(PFRED "packet-helper Has no eth0 ip address %d\n" PFNRM, time(0));		
	exit(-1);
}
//mabye add hash table old ip:port -> new ip:port

int CheckAndTransmit(int readSocket, int writeSock, struct timeval timeout)
{
	//check socket for incomming data
	fd_set sockSet;
	FD_ZERO(&sockSet);
	FD_SET(readSocket, &sockSet);
	int sv = select(readSocket + 1, &sockSet, NULL, NULL, &timeout);
	
	if (sv < 0)
		return sv;//An error occured
	else if (sv == 0)
		return 0;//timeout occured
	
	const int buflen = 65566;
	char buffer[buflen];
	memset(buffer, 0, sizeof(buffer));

	int readlen = read(readSocket, buffer, buflen);
	if (readlen < 0)
		return readlen;//Read error occured

	int writelen = write(writeSock, buffer, readlen);
	if (writelen < 0)
		return writelen;//Write error occured

	return 0;
}

void* ThreadLink(void *data)
{
	linkInit link = ((linkInit*)data)[0];

	char clientIPString[INET6_ADDRSTRLEN];
	ipv4ToString(link.addr.sin_addr.s_addr, clientIPString);
	printf(PFGRN "packet-helper Started Link Management Thread for IP:%s %d\n" PFNRM, clientIPString, time(0));

	char serverIPString[INET6_ADDRSTRLEN];
	int clientSocket, serverSock;
	clientSocket = link.socket;
	struct sockaddr_in servaddr, clientaddr;
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = config.ipAddressOut;
	servaddr.sin_port = htons(config.portOut);
	clientaddr = link.addr;
	ipv4ToString(config.ipAddressOut, serverIPString);

	if (!(serverSock = socket(AF_INET, SOCK_STREAM, 0))) { printf(PFRED "creating link server socket failed %s %d\n" PFNRM, serverIPString, time(0)); exit(-1); }
	if (connect(serverSock, (struct sockaddr*)&servaddr, sizeof(struct sockaddr)) != 0) { printf(PFRED "connecting to server failed %s %d\n" PFNRM, serverIPString, time(0)); exit(-1); }

	struct timeval timeout;
	timeout.tv_sec = 0;
	timeout.tv_usec = 500; //0.5ms timeout

	int lastActiveTime = time(0);
	linkCount++;

	while (1)
	{
		int client2Server = 0, server2Client = 0;
		client2Server = CheckAndTransmit(clientSocket, serverSock, timeout);
		server2Client = CheckAndTransmit(serverSock, clientSocket, timeout);
		if (client2Server < 0 || server2Client < 0)
		{
			printf(PFYEL "something happened a=%d b=%d closing for IP:%s %d\n" PFNRM, client2Server, server2Client, clientIPString, time(0));
			break;
		}
		else if (client2Server > 0 || server2Client > 0)
		{
			lastActiveTime = time(0);
		}

		sleep(0);

		if (time(0) - lastActiveTime > linkTimeout)
		{
			printf(PFYEL "We have not recieved any messages in %ds timing outfor IP:%s %d\n" PFNRM, linkTimeout, clientIPString, time(0));
			break;
		}
	}
	
	printf(PFGRN "packet-helper Ended Link Management Thread for IP:%s %d\n" PFNRM, clientIPString, time(0));

	linkCount--;
	close(clientSocket);
	close(serverSock);
	pthread_detach(pthread_self());

	return NULL;
}

void* ThreadRouterMaster(void *data)
{
	int inSocket, clientSock;
	struct sockaddr_in servaddr, clientaddr;
	inSocket = ((int*)data)[0];
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port = htons(config.portIn);

	signal(SIGPIPE, SIG_IGN);

	if (bind(inSocket, (struct sockaddr*)&servaddr, sizeof(struct sockaddr)) <  0) { printf(PFRED "binding router socket failed\n" PFNRM); exit(-1); }
	if (listen(inSocket, 10) < 0) { printf(PFRED "router socket listen failed\n" PFNRM); exit(-1); }

	printf(PFGRN "packet-helper Router Socket Ready %d\n" PFNRM, time(0));

	while(1)
	{
		socklen_t clientLen = sizeof(clientaddr);
		clientSock = accept(inSocket, (struct sockaddr*)&clientaddr, &clientLen);

		if (!AddressIsGood(clientaddr, clientLen))
		{
			
			char ipString[INET6_ADDRSTRLEN];
			ipv4ToString(clientaddr.sin_addr.s_addr, ipString);
			printf(PFRED "Recieved message from invalid client %s %d\n" PFNRM, ipString, time(0));
			close(clientSock);
		}
		else
		{
			linkInit link;
			link.addrlen = clientLen;
			link.addr = clientaddr;
			link.socket = clientSock;

			pthread_t linkThread;
			if (pthread_create(&linkThread, NULL, &ThreadLink, (void*)&link) != 0)
			{
				printf(PFRED "router failed to create child thread\n" PFNRM);
				close(clientSock);
			}
		}
		
		sleep(1);
	}
}

void* ThreadEnabler(void *data)
{
	int inSocket, clientSock;
	struct sockaddr_in servaddr, clientaddr;
	inSocket = ((int*)data)[0];
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port = htons(config.portUpdate);

	signal(SIGPIPE, SIG_IGN);

	if (bind(inSocket, (struct sockaddr*)&servaddr, sizeof(struct sockaddr)) <  0) { printf(PFRED "binding enabler socket failed\n" PFNRM); exit(-1); }
	if (listen(inSocket, 10) < 0) { printf(PFRED "enabler socket listen failed\n" PFNRM); exit(-1); }

	printf(PFGRN "packet-helper Enabler Socket Ready %d\n" PFNRM, time(0));

	while(1)
	{
		socklen_t clientLen = sizeof(clientaddr);
		clientSock = accept(inSocket, (struct sockaddr*)&clientaddr, &clientLen);

		char ipString[INET6_ADDRSTRLEN];
		ipv4ToString(clientaddr.sin_addr.s_addr, ipString);

		const int buflen = 65566;
		char buffer[buflen];
		memset(buffer, 0, sizeof(buffer));

		int readlen = read(clientSock, buffer, buflen);
		if (readlen < 0)
		{
			printf(PFRED "Recieved corrupt request IP:%s %d\n" PFNRM, ipString, time(0));
		}
		else
		{
			if (strstr(buffer, ACCESSKEY) > 0)
			{
				printf(PFGRN "Recieved enable request with key from from Client:%s %d\n" PFNRM, ipString, time(0));
				int index = tempIndex++ % 10;
				tempAddresses[index].ipAddress = clientaddr.sin_addr.s_addr;
				tempAddresses[index].endOfLife = time(0) + 60;
				write(clientSock, HTTP_AFFIRM, sizeof(HTTP_AFFIRM));
			}
			else
			{
				printf(PFRED "Recieved invalid request IP:%s %d\n" PFNRM, ipString, time(0));
			}
		}
		close(clientSock);
		
		sleep(1);
	}
}


int main(void)
{
	int itr, error;
	printf(PFGRN "packet-helper Started %d\n" PFNRM, time(0));

	if (error = ReadConfig(&config))
		return -1;

	char lastIPstr[INET6_ADDRSTRLEN];
	ipv4ToString(config.ipAddressList[config.ipAddressCount-1], lastIPstr);

	printf(PFGRN "packet-helper Loaded Config in=%d out=%d up=%d IPcount=%d lastIP=%s %d\n" PFNRM, 
		config.portIn, config.portOut, config.portUpdate, config.ipAddressCount, lastIPstr, time(0));

	getMyIPAddress();

	int routerInSocket, enableInSocket;
	routerInSocket = socket(AF_INET, SOCK_STREAM, 0);//SOCK_STREAM = tcp	//SOCK_DGRAM = udp
	enableInSocket = socket(AF_INET, SOCK_STREAM, 0);

	pthread_t routerThread, enableThread;
	int threadErr = 0;
	threadErr |= pthread_create(&routerThread, NULL, &ThreadRouterMaster, (void*)&routerInSocket);
	threadErr |= pthread_create(&enableThread, NULL, &ThreadEnabler, (void*)&enableInSocket);

	if (threadErr)
	{
		printf(PFRED "packet-helper Thread Creation error %d\n" PFNRM, time(0));
		return -1;
	}

	char typeString[100];
	while(1)
	{
		scanf("%99s", typeString);

		if (strcmp(typeString, "quit") == 0)
		{
			printf(PFCYN "packet-helper Quitting as requested %d\n" PFNRM, time(0));
			close(routerInSocket);
			close(enableInSocket);
			pthread_detach(routerThread);
			pthread_detach(enableThread);
			return 0;
		}
		else if (strcmp(typeString, "hi") == 0)
		{
			printf(PFCYN "packet-helper says hello %d\n" PFNRM, time(0));
		}
		else if (strcmp(typeString, "link") == 0 || strcmp(typeString, "links") == 0)
		{
			printf(PFCYN "packet-helper says there are %d links active %d\n" PFNRM, linkCount, time(0));
		}

		sleep(1);
	}

}